"""
   Copyright 2019-2020 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import logging
from argparse import ArgumentParser, Namespace
from time import time
from typing import Tuple

import numpy as np
import torch
from ignite.engine import Engine, Events, create_supervised_trainer
from ignite.handlers import EarlyStopping
from ignite.metrics.loss import Loss
from torch.nn import Module
from torch.optim import Adam
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from neural_semigroups.cayley_database import CayleyDatabase
from neural_semigroups.magma import Magma


def load_database_as_cubes(
        database_filename: str,
        train_size: int,
        validation_size: int
) -> Tuple[
    np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray
]:
    """
    load a database file to probability cubes representation

    :param database_filename: the name of the file from which to extract data
    :param train_size: number of tables for training
    :param validation_size: number of tables for validation
    :returns: three arrays of probability Cayley cubes (train, validation, test
    ) and three arrays of labels for them
    """
    cayley_db = CayleyDatabase()
    logging.info("reading data from disk")
    try:
        cayley_db.load_database(database_filename)
    except ValueError:
        cayley_db.load_smallsemi_database(database_filename)
    logging.info("splitting by train and test")
    train, validation, test = cayley_db.train_test_split(
        train_size, validation_size
    )
    logging.info("generating train cubes")
    train_cubes = list()
    for cayley_table in tqdm(train.database):
        train_cubes.append(Magma(cayley_table).probabilistic_cube)
    validation_cubes = list()
    logging.info("generating validation cubes")
    for cayley_table in tqdm(validation.database):
        validation_cubes.append(Magma(cayley_table).probabilistic_cube)
    test_cubes = list()
    logging.info("generating test cubes")
    for cayley_table in tqdm(test.database):
        test_cubes.append(Magma(cayley_table).probabilistic_cube)
    return (
        np.stack(train_cubes), np.stack(validation_cubes),
        np.stack(test_cubes), train.labels, validation.labels, test.labels
    )


def get_arguments() -> Namespace:
    """
    parse script arguments

    :returns: script parameters
    """
    parser = ArgumentParser()
    parser.add_argument(
        "--cardinality",
        type=int,
        help="magma cardinality",
        required=True,
        choices=range(2, 8)
    )
    parser.add_argument(
        "--epochs",
        type=int,
        help="number of epochs to train",
        default=100,
        required=False
    )
    parser.add_argument(
        "--learning_rate",
        type=float,
        help="learning rate",
        default=0.001,
        required=False
    )
    parser.add_argument(
        "--weight_decay",
        type=float,
        help="weight decay",
        default=0.0,
        required=False
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        help="batch size for training",
        default=32,
        required=False
    )
    parser.add_argument(
        "--train_size",
        type=int,
        help="number of tables for training",
        required=True
    )
    parser.add_argument(
        "--validation_size",
        type=int,
        help="number of tables for validation",
        required=True
    )
    return parser.parse_args()


def learning_pipeline(
        args: Namespace,
        model: Module,
        evaluator: Engine,
        loss: Loss,
        data_loaders: Tuple[DataLoader, DataLoader],
) -> None:
    """
    run a comon learning pipeline

    :param args: some other arguments from the shell script
    :param model: a network architecture
    :param evaluator: an `ignite` engine which evaluates the model's quality
    :param loss: the criterion to optimize
    :oaram data_loader: train and validation data loaders
    """
    logging.info("data prepared")
    optimizer = Adam(
        model.parameters(),
        lr=args.learning_rate,
        weight_decay=args.weight_decay
    )
    trainer = create_supervised_trainer(model, optimizer, loss)

    def score_function(engine):
        val_loss = engine.state.metrics["loss"]
        return -val_loss
    handler = EarlyStopping(
        patience=10, score_function=score_function, trainer=trainer
    )
    evaluator.add_event_handler(Events.COMPLETED, handler)
    writer = SummaryWriter()
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    @trainer.on(Events.EPOCH_COMPLETED)
    # pylint: disable=unused-variable
    def log_training_results(trainer):
        evaluator.run(data_loaders[0])
        # pylint: disable=no-member
        for metric in evaluator.state.metrics.keys():
            writer.add_scalars(
                metric,
                # pylint: disable=no-member
                {"training": evaluator.state.metrics[metric]},
                global_step=trainer.state.iteration,
                walltime=int(time())
            )

    @trainer.on(Events.EPOCH_COMPLETED)
    # pylint: disable=unused-variable
    def log_validation_results(trainer):
        evaluator.run(data_loaders[1])
        # pylint: disable=no-member
        for metric in evaluator.state.metrics.keys():
            writer.add_scalars(
                metric,
                # pylint: disable=no-member
                {"validation": evaluator.state.metrics[metric]},
                global_step=trainer.state.iteration,
                walltime=int(time())
            )
        # pylint: disable=no-member
        print(evaluator.state.metrics["loss"])
        torch.save(model, f"semigroups.{args.cardinality}.model")
    logging.info("training started")
    trainer.run(data_loaders[0], max_epochs=args.epochs)


def get_loaders(
        database_filename: str,
        batch_size: int,
        train_size: int,
        validation_size: int,
        use_labels: bool
) -> Tuple[DataLoader, DataLoader]:
    """
    get train and validation data loaders

    :param database_filename: the name of the file from which to extract data
    :param batch_size: batch size (common for train and validation)
    :param train_size: number of tables for training
    :param validation_size: number of tables for validation
    :param use_labels: whether to set a target as labels from database (for
    classifier) or to use X's as labels (for autoencoder)
    :returns: a pair of train and validation data loaders
    """
    (
        train, validation, _,
        train_labels, validation_labels, _
    ) = load_database_as_cubes(
        database_filename, train_size, validation_size
    )
    train_tensor = torch.from_numpy(train)
    val_tensor = torch.from_numpy(validation)
    if use_labels:
        train_data = TensorDataset(
            train_tensor, torch.from_numpy(train_labels)
        )
        val_data = TensorDataset(
            val_tensor, torch.from_numpy(validation_labels)
        )
    else:
        train_data = TensorDataset(train_tensor, train_tensor)
        val_data = TensorDataset(val_tensor, val_tensor)
    return (
        DataLoader(train_data, batch_size=batch_size, shuffle=True),
        DataLoader(val_data, batch_size=batch_size, shuffle=True)
    )



"""
   Copyright 2019-2020 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from os.path import basename
from typing import List, Tuple

import numpy as np
import torch
from torch.nn import Module
from tqdm import tqdm

from neural_semigroups.utils import (check_filename, check_smallsemi_filename,
                                     get_equivalent_magmas,
                                     import_smallsemi_format)


class CayleyDatabase:
    """
    a database of Cayley tables with different utility functions
    """
    # the number of elements in an underlying magma
    cardinality: int
    # a database of known Cayley tables of shape [:, cardinality, cardinality]
    database: np.ndarray
    # a pre-trained PyTorch Model
    model: Module
    # 1-D labels array of the same length as the database
    labels: np.ndarray

    def load_database(self, filename: str) -> None:
        """
        loads a known Cayley tables database from a file
        Database file description:

        * a filename is of a form [description].[n].npz
        * a file is of ``npz`` format readable by numpy

        :param filename: where to load a database from
        :returns:
        """
        self.cardinality = check_filename(basename(filename))
        npz_file = np.load(filename)
        self.database = npz_file["database"]
        self.labels = npz_file.get(
            "labels", np.zeros(len(self.database), dtype=np.int64)
        )
        npz_file.close()

    def load_smallsemi_database(self, filename: str) -> None:
        """
        loads a known Cayley tables database from a file in ``smallsemi`` format
        Format description:

        * filename is of a form ``data[n].gl``, :math:`1<=n<=7`
        * lines are separated by a pair of symbols ``\\r\\n``
        * there are exactly :math:`n^2` lines in a file
        * the first line is a header starting with '#' symbol
        * each line is a string of :math:`N` digits from :math:`0` to :math:`n-1`
        * :math`N` is the number of semigroups in the database
        * each column represents a serialised Cayley table
        * the database contains only cells starting from the second
        * the first cell of each Cayley table is assumed to be filled with ``0``

        :param filename: where to load a database from
        :returns:
        """
        self.cardinality = check_smallsemi_filename(basename(filename))
        with open(filename, "r") as database:
            self.database = import_smallsemi_format(database.readlines())
        self.labels = np.ones(len(self.database), dtype=np.int64)

    def augment_by_equivalent_tables(self) -> None:
        """
        for every Cayley table in a previously loaded database adds all of its
        equivalent tables to the database
        """
        database: List[np.ndarray] = []
        for table in tqdm(self.database):
            database += [get_equivalent_magmas(table)]
        self.database = np.unique(np.concatenate(database, axis=0), axis=0)

    def _check_input(self, cayley_table: List[List[int]]) -> bool:
        """
        checks the input to be a correct Cayley table

        :param cayley_table: a partially filled Cayley table (unknow entries are filled by ``-1``)
        :returns: whether the input is correct
        """
        correct = True
        table = np.array(cayley_table)
        if table.shape != (self.cardinality, self.cardinality):
            correct = False
        elif table.dtype != int:
            correct = False
        elif table.max() >= self.cardinality or table.min() < -1:
            correct = False
        return correct

    def search_database(
            self,
            cayley_table: List[List[int]]
    ) -> List[np.ndarray]:
        """
        get a list of possible completions of a partially filled Cayley table (unknow entries are filled by ``-1``)

        :param cayley_table: a partially filled Cayley table (unknow entries are filled by ``-1``)
        :returns: a list of Cayley tables
        """
        if not self._check_input(cayley_table):
            raise ValueError(
                f"invalid Cayley table of {self.cardinality} elements"
            )
        completions = list()
        if self.database is not None:
            partial_table = np.array(cayley_table)
            rows, cols = np.where(partial_table != -1)
            for table in tqdm(self.database):
                if np.alltrue(table[rows, cols] == partial_table[rows, cols]):
                    completions.append(table)
        return completions

    def fill_in_with_model(
            self,
            cayley_table: List[List[int]]
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        get a list of possible completions of a partially filled Cayley
        table (unknow entries are filled by ``-1``) using a machine learning model

        :param cayley_table: a partially filled Cayley table (unknow entries are filled by ``-1``)
        :returns: a tuple: (most probable completion, probabilistic cube)
        """
        if not self._check_input(cayley_table):
            raise ValueError(
                f"invalid Cayley table of {self.cardinality} elements"
            )
        table = np.array(cayley_table)
        inv_cardinality = 1 / self.cardinality
        cube = np.zeros(
            [self.cardinality, self.cardinality, self.cardinality],
            dtype=np.float32
        )
        rows, cols = np.where(table != -1)
        cube[rows, cols, table[rows, cols]] = 1.0
        rows, cols = np.where(table == -1)
        cube[rows, cols, :] = inv_cardinality
        prediction = self.model(torch.from_numpy(
            cube.reshape([
                -1, self.cardinality, self.cardinality, self.cardinality
            ])
        )).detach().numpy()[0]
        return (prediction.argmax(axis=-1), prediction)

    def load_model(self, filename: str) -> None:
        """
        load pre-trained PyTorch model

        :param filename: where to load the model from
        :returns:
        """
        self.model = torch.load(filename)

    def train_test_split(
            self,
            train_size: int,
            validation_size: int
    ) -> Tuple["CayleyDatabase", "CayleyDatabase", "CayleyDatabase"]:
        """
        split a database of Cayley table in three: train, validation, and test

        :param cayley_db: a database of Cayley tables
        :param train_size: number of tables in a train set
        :param train_size: number of tables in a validation set
        :returns: a triple of distinct Cayley tables databases: ``(train, validation, test)``
        """
        all_indices = np.arange(len(self.database))
        np.random.shuffle(all_indices)
        train_indices = all_indices[:train_size]
        validation_indices = all_indices[
            train_size:train_size + validation_size
        ]
        test_indices = all_indices[train_size + validation_size:]
        train = CayleyDatabase()
        train.database = self.database[train_indices]
        train.labels = self.labels[train_indices]
        validation = CayleyDatabase()
        validation.database = self.database[validation_indices]
        validation.labels = self.labels[validation_indices]
        test = CayleyDatabase()
        test.database = self.database[test_indices]
        test.labels = self.labels[test_indices]
        return train, validation, test



"""
   Copyright 2019-2020 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from collections import OrderedDict
from typing import List, no_type_check

from torch import Tensor
from torch.nn import BatchNorm1d, Linear, Module, ReLU, Sequential, Sigmoid


class MagmaClassifier(Module):
    """
    Feed Forward Network for probability Cayley cubes of magmas
    """
    def __init__(
            self,
            cardinality: int,
            hidden_dims: List[int]
    ):
        """
        :param cardinality: the number of elements in a magma
        :param hidden_dims: a list of sizes of hidden layers of a feed-forward network
        """
        super().__init__()
        self.cardinality = cardinality
        self.input_dim = cardinality ** 3
        hidden_layers: "OrderedDict[str, Module]" = OrderedDict()
        dims = [self.input_dim] + hidden_dims
        for i in range(len(dims) - 1):
            hidden_layers.update(
                {f"linear{i}": Linear(dims[i], dims[i + 1], bias=True)}
            )
            hidden_layers.update({f"relu{i}": ReLU()})
            hidden_layers.update({f"bn{i}": BatchNorm1d(dims[i + 1])})
        self.hidden_layers = Sequential(hidden_layers)
        top_layers: "OrderedDict[str, Module]" = OrderedDict()
        top_layers.update(
            {f"linear{len(dims)}": Linear(dims[-1], 2, bias=False)}
        )
        top_layers.update({f"sigmoid": Sigmoid()})
        self.top_layers = Sequential(top_layers)

    # pylint: disable=arguments-differ
    @no_type_check
    def forward(self, cayley_cube: Tensor) -> Tensor:
        """
        forward pass inhereted from Module

        :param cayley_cube: probabilistic representation of a magma
        :returns: score of associativity (from ``0`` to ``1``)
        """
        last_hidden_layer = self.hidden_layers(
            cayley_cube.view(-1, self.input_dim)
        )
        return self.top_layers(last_hidden_layer)



"""
   Copyright 2019-2020 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from itertools import permutations
from typing import List, Tuple

import numpy as np

from neural_semigroups.magma import Magma

# the Cayley table of Klein Vierergruppe
FOUR_GROUP = np.array([
    [0, 1, 2, 3],
    [1, 0, 3, 2],
    [2, 3, 0, 1],
    [3, 2, 1, 0]
])

# some non associative magma
# (0 * 1) * 2 = 0 * 2 = 0
# 0 * (1 * 2) = 0 * 0 = 1
NON_ASSOCIATIVE_MAGMA = np.array([
    [1, 0, 0], [2, 2, 0], [2, 2, 2]
])


def random_magma(cardinality: int) -> np.ndarray:
    """
    randomly generate a Cayley table for a magma

    :param cardinality: number of elements in a magma
    :returns: Cayley table of some magma
    """
    return np.random.randint(
        low=0,
        high=cardinality,
        size=cardinality * cardinality
    ).reshape(cardinality, cardinality)


def random_semigroup(dim: int, maximal_tries: int) -> Tuple[bool, np.ndarray]:
    """
    randomly serch for a semigroup Cayley table.
    Not recommended to use with dim > 4

    :param dim: number of elements in a semigroup
    :param maximal_tries: how many times to try at most
    :returns: a pair (whether the Cayley table is associative,
    a Cayley table of a magma)
    """
    associative = False
    try_count = 0
    while not associative and try_count <= maximal_tries:
        mult = random_magma(dim)
        associative = Magma(mult).is_associative
        try_count += 1
    return associative, mult


def next_magma(magma: Magma) -> Magma:
    """
    goes to the next magma Cayley table in their lexicographical
    order

    :param magma: a magma
    :returns: a magma
    """
    next_table = magma.cayley_table.copy()
    one = 1
    row = magma.cardinality - 1
    column = magma.cardinality - 1
    while one == 1:
        if next_table[row, column] < magma.cardinality - 1:
            next_table[row, column] += 1
            one = 0
        else:
            if column > 0:
                next_table[row, column] = 0
                column -= 1
            else:
                if row > 0:
                    next_table[row, column] = 0
                    row -= 1
                    column = magma.cardinality - 1
                else:
                    raise Exception("there is no next magma!")
    return Magma(next_table)


def check_filename(filename: str) -> int:
    """
    checks filename, raises if it's incorrect

    :param filename: filename to check
    :returns: magma cardinality extracted from the filename
    """
    wrong_name = False
    if not isinstance(filename, str):
        wrong_name = True
    else:
        filename_parts = filename.split(".")
        if len(filename_parts) != 3:
            wrong_name = True
        elif filename_parts[0] not in ("semigroup", "monoid", "group"):
            wrong_name = True
        elif filename_parts[2] != "npz":
            wrong_name = True
        elif not filename_parts[1].isdigit():
            wrong_name = True
        else:
            cardinality = int(filename_parts[1])
    if wrong_name:
        raise ValueError(
            f"""filename should be of format
[semigroup|monoid|group].[int].npz, not {filename}"""
        )
    return cardinality


def check_smallsemi_filename(filename: str) -> int:
    """
    checks a filename from a `smallsemi` package, raises if it's incorrect

    :param filename: filename from a `smallsemi` package to check
    :returns: magma cardinality extracted from the filename
    """
    wrong_name = False
    if not isinstance(filename, str):
        wrong_name = True
    else:
        filename_parts = filename.split(".")
        if len(filename_parts) != 2:
            wrong_name = True
        elif filename_parts[1] != "gl":
            wrong_name = True
        elif filename_parts[0][:-1] != "data":
            wrong_name = True
        elif not filename_parts[0][-1].isdigit():
            wrong_name = True
        else:
            cardinality = int(filename_parts[0][-1])
    if wrong_name:
        raise ValueError(
            f"filename should be of format data[1-7].gl, not {filename}"
        )
    return cardinality


def get_magma_by_index(cardinality: int, index: int) -> Magma:
    """
    find a magma from a lexicographical order by its index

    :param cardinality: the number of elements in a magma
    :param index: an index of magma in a lexicographical order
    :returns: a magma with a given index
    """
    square = cardinality ** 2
    if index < 0 or index >= cardinality ** square:
        raise ValueError("""
        An index must be non negative and less than $n^(n^2)$""")
    cayley_table = list()
    residual = index
    for _ in range(square):
        cayley_table.append(residual % cardinality)
        residual = residual // cardinality
    return Magma(np.array(
        list(reversed(cayley_table))
    ).reshape(cardinality, cardinality))


def import_smallsemi_format(lines: List[str]) -> np.ndarray:
    """
    imports lines in a format used by `smallsemi` GAP package
    https://www.gap-system.org/Manuals/pkg/smallsemi-0.6.11/doc/chap0.html

    :param lines: lines read from a file of `smallsemi` format
    :returns: a list of Cayley tables
    """
    raw_tables = np.array(
        [list(map(int, list(line[:-1]))) for line in lines[1:]]
    ).T
    tables = np.hstack([
        np.zeros([raw_tables.shape[0], 1], dtype=int),
        raw_tables
    ])
    cardinality = int(tables.max()) + 1
    return tables.reshape(tables.shape[0], cardinality, cardinality)


def get_isomorphic_magmas(cayley_table: np.ndarray) -> np.ndarray:
    """
    given a Cayley table of a magma generate Cayley tables of isomorphic magmas
    by appying all possible permutations of the magma's elements

    :param cayley_table: a Cayley table of a magma
    :returns: a list of Cayley tables of isomorphic magmas
    """
    isomorphic_cayley_tables = list()
    dim = cayley_table.shape[0]
    for permutation in permutations(range(dim)):
        isomorphic_cayley_table = np.array(np.zeros_like(
            cayley_table, dtype=int
        ))
        for i in range(dim):
            for j in range(dim):
                isomorphic_cayley_table[
                    permutation[i], permutation[j]
                ] = permutation[cayley_table[i, j]]
        isomorphic_cayley_tables.append(isomorphic_cayley_table)
    return np.unique(np.array(isomorphic_cayley_tables), axis=0)


def get_anti_isomorphic_magmas(cayley_table: np.ndarray) -> np.ndarray:
    """
    given a Cayley table of a magma generate Cayley tables of anti-isomorphic
    magmas by appying all possible permutations of the magma's elements

    :param cayley_table: a Cayley table of a magma
    :returns: a list of Cayley tables of anti-isomorphic magmas
    """
    anti_isomorphic_cayley_tables = list()
    dim = cayley_table.shape[0]
    for permutation in permutations(range(dim)):
        anti_isomorphic_cayley_table = np.array(np.zeros_like(
            cayley_table, dtype=int
        ))
        for i in range(dim):
            for j in range(dim):
                anti_isomorphic_cayley_table[
                    permutation[i], permutation[j]
                ] = permutation[cayley_table[j, i]]
        anti_isomorphic_cayley_tables.append(anti_isomorphic_cayley_table)
    return np.unique(np.array(anti_isomorphic_cayley_tables), axis=0)


def get_equivalent_magmas(cayley_table: np.ndarray) -> np.ndarray:
    """
    given a Cayley table of a magma generate Cayley tables of isomorphic and
    anti-isomorphic magmas

    :param cayley_table: a Cayley table of a magma
    :returns: a list of Cayley tables of isomorphic and anti-isomorphic magmas
    """
    equivalent_tables = np.concatenate([
        get_isomorphic_magmas(cayley_table),
        get_anti_isomorphic_magmas(cayley_table)
    ], axis=0)
    return np.unique(equivalent_tables, axis=0)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neural_semigroups']

package_data = \
{'': ['*']}

install_requires = \
['numpy==1.18.1',
 'pandas==1.0.2',
 'pytorch-ignite==0.3.0',
 'sphinx_rtd_theme==0.4.3',
 'tensorboard==2.1.1',
 'torch==1.4.0',
 'tqdm==4.43.0']

setup_kwargs = {
    'name': 'neural-semigroups',
    'version': '0.2.1',
    'description': 'Neural networks powered research of semigroups',
    'long_description': "# Neural Semigroups\n\nHere we try to learn Cayley tables of semigroups using neural\nnetworks. The supposed workflow:\n\n* install the package\n* get the data\n* build a model\n* print a model testing report\n\nMore documentation can be found [here](https://neural-semigroups.readthedocs.io).\n\n## Motivation\n\nThis work was inspired by [a sudoku\nsolver](https://github.com/Kyubyong/sudoku). A solved Sudoku puzzle\nis nothing more than a Cayley table of a quasigroup from 9 items with\nsome well-known additional properties. So, one can imagine a puzzle\nmade from a Cayley table of any other magma, e. g. a semigroup, by\nhiding part of its cells.\n\nThere are two major differences between sudoku and puzzles based on\nsemigroups:\n\n1) it's easy to take a glance on a table to understand whether it is\na sudoku or not. That's why it was possible to encode numbers in a\ntable cells as colour intensities. Sudoku is a picture, and a\nsemigroup is not. It's difficult to check a Cayley table's\nassociativity with a naked eye;\n\n2) sudoku puzzles are solved by humans for fun and thus catalogued.\nWhen solving a sudoku one knows for sure that there is a unique\nsolution. On the contrary, nobody guesses values in a partially\nfilled Cayley table of a semigroup as a form of amuzement. As a\nresult, one can create a puzzle from a full Cayley table of a\nsemigroup but there may be many distinct solutions.\n\n## Package Installation\n\nFirst, get Python 3.8\n\nThen run several commands in a terminal:\n\n```bash\ngit clone git@bitbucket.org:inpefess/neural-semigroups.git\ncd neural-semigroups\npython -m venv venv\nsource ./venv/bin/activate\npip install -U pip poetry\npoetry install\n```\n\nThis could take some time (mostly downloading `torch`).\n\nAfter that you can go to the project's directory and `source\n./venv/bin/activate` to start a virtual environment there.\n\n## Getting Data\n\nThis package uses data included into a `smallsemi` package for GAP\nsystem. One can download it from [the GAP\npage](https://www.gap-system.org/pub/gap/gap4/tar.gz/packages/smallsemi-0.6.11.tar.gz).\n\nYou can get the data by running a script from a `scripts` folder:\n```bash\ncd scripts\n./download_smallsemi.sh\n```\n\n## Training a Model\n\nHere are several examples of commands to train a model depending on\nsemigroup's cardinality:\n\n```bash\npython train_denoising_autoencoder.py --cardinality 4 --epochs 100 \\\n--learning_rate 0.1 --batch_size 32 --train_size 10 --validation_size 10\n```\n```bash\npython train_denoising_autoencoder.py --cardinality 5 --epochs 100 \\\n--learning_rate 0.01 --batch_size 256 --train_size 100 --validation_size 100\n```\n```bash\npython train_denoising_autoencoder.py --cardinality 6 --epochs 100 \\\n--learning_rate 0.001 --batch_size 2048 --train_size 1000 --validation_size 100\n```\n```bash\npython train_denoising_autoencoder.py --cardinality 7 --epochs 100 \\\n--learning_rate 0.001 --batch_size 2048 --train_size 1000 --validation_size 100\n```\n\n## Printing a Testing Report\n\nOne can print a model quality report using the following command:\n\n```bash\npython test_model.py --cardinality 4\n```\n",
    'author': 'Boris Shminke',
    'author_email': 'boris@shminke.ml',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/inpefess/neural-semigroups',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
   Copyright 2019-2020 Boris Shminke

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from neural_semigroups.associator_loss import AssociatorLoss
from neural_semigroups.cayley_database import CayleyDatabase
from neural_semigroups.classifier import MagmaClassifier
from neural_semigroups.cyclic_group import CyclicGroup
from neural_semigroups.denoising_autoencoder import MagmaDAE
from neural_semigroups.magma import Magma
